/*
Lexmark 1100 printer driver for Linux
Copyright (C) 1999   Tim Engler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "Head.h"
#include "Coord.h"
#include "math.h"
#include "errors.h"
#include <limits.h>

Head::Head(Map *imageMap, LexmarkProtocol *protocol, Rect *pageDimensions)
{
	this->jets = protocol->createJets(&this->totalJets,imageMap);
	
	// adjust the range based on the position of each jet
	for(int i = 0; i < totalJets; i++)
	{
		Rect jetRange;
		
		jets[i]->getRange(&jetRange);
		headRange.getMax(&headRange,&jetRange);
	}

	this->imageMap = imageMap;
	this->protocol = protocol;
	this->pageDimensions = pageDimensions;
}

Head::~Head()
{
	for(int i = 0; i < totalJets; i++)
		delete jets[i];

	//delete the array itself
	delete jets;
}


//returns true if the image fits nicely in the print zone
bool Head::imageFits()
{
	if(imageMap->width > pageDimensions->width - protocol->getLeftBorder() - 
	   protocol->getRightBorder() ||
	   imageMap->height > pageDimensions->height - protocol->getTopBorder() - 
	   protocol->getBottomBorder())
		return false; // image doesn't fit
	
	return true;
}

//adjust the starting x and y positions for the granualarity of the protocol and print zone
void Head::adjustPositions(int *startX, int *y, int *endX)
{
	protocol->adjustPositions(startX,y,endX);
	
	//make sure image isn't outside of the print zone in x
	if(*endX > this->pageDimensions->width - protocol->getLeftBorder())
		*endX = this->pageDimensions->width - protocol->getRightBorder();
	if(*startX > this->pageDimensions->width - protocol->getRightBorder())
		die("Image off the page");
	if(*startX < protocol->getLeftBorder())
		*startX = protocol->getLeftBorder();

	//make sure image isn't outside of the print zone in y
	if(*y < protocol->getTopBorder())
		*y = protocol->getTopBorder();
}

void Head::paint()
{	
    int first = 1;
	int i;

	Coord imageStart(0,0);
	Coord jetStart(0,0);

	//if the image fits inside the printable area, we will adjust it so we print the whole thing
	//(generic image mode)
	if(imageFits())
	{
		//ask each jet for it's ideal starting point for the image and take the lowest common denominator(farthest down and right)
		for(i = 0; i < totalJets; i++)
		{
			jets[i]->getPreferredImageStart(&jetStart);

			imageStart.max(&imageStart,&jetStart);
		}
		
		//add the left and top borders from the protocol
		imageStart.x += protocol->getLeftBorder();
		imageStart.y += protocol->getTopBorder();
		
	}
	else // the image is too big for the printable area of the page, center it on the page for us
	{
		imageStart.x = (this->pageDimensions->height - imageMap->height ) / 2;
		imageStart.y = (this->pageDimensions->width - imageMap->width ) / 2;
	}

	Rect range; // range of area left for the head to cover
	
	for(i = 0; i < totalJets; i++)
	{
		Rect jetRange;
		
		//tell each jet the chosen starting position
		jets[i]->setImageStart(&imageStart);

		//ask it for the range to cover the image, we will increase the range to cover all jet positions
		jets[i]->getImageRange(&jetRange);
		
		range.getMax(&range,&jetRange);
	}

	if(range.height + range.y > pageDimensions->height - protocol->getBottomBorder())
		range.height = pageDimensions->height - protocol->getBottomBorder()- range.y;

    //start the page
    protocol->startPage();

    //paint the image
    while(1)
    {
		int finished;

		int y;

		//first we need to find the next actual spot we need to paint on
		//we have the head tell each jet which point to examine, so it can coordinate caching the image.
		//If we had each jet skip the blank lines themselves, they each might need to go through half the image, and that
		//would be a waste of memory
		for(y = range.y;y < range.y + range.height; y ++)
		{
			//cache the area covering all the jets
			imageMap->cacheLines(headRange.y+y - imageStart.y,headRange.height);
			
			int j;
	
			for(j = 0; j < totalJets; j ++)
			{
				if(jets[j]->needsCover(y))
					break; // this will be the starting y point of the paint
			}

			if(j != totalJets)
				break; // at least one jet needed to paint on the current line

		}

		if(y == range.y + range.height) //finished
		{
			//eject page
			protocol->endPage();
			break;
		}

		range.height -= y - range.y;
		range.y = y; // this will recover some already viewed data, but the Jets will properly ignore it. They
		// know what they have covered

		int startX= INT_MAX, endX = INT_MIN;

		//now make each jet calculate it's startX and endX for this line
		for(int j = 0; j < totalJets; j++)
		{
			int jetStartX, jetEndX;
		
			if(jets[j]->findStartEnd(y,&jetStartX,&jetEndX)) // if the jet needs to cover the area at all
			{
				startX = min(startX,jetStartX);
				endX = max(endX,jetEndX);
			}
			
		}

		//ask protocol to adjust the start pos if necessary
		protocol->adjustPositions(&startX, &y, &endX);

		//paint!

		//move head into position
		protocol->startPaint(startX, y,endX-startX);

		//get information from each jet for this stripe
		for(i=0; i < totalJets; i++)
		{
			jets[i]->paint(startX, endX, y, protocol);
		}

		protocol->endPaint();
	}
}
